@@ -1,20 +1,71 @@
 #############################################################################
 # Grid-management and layout preperation.
 #
-# (c) by Tels 2004-2005.
+# (c) by Tels 2004-2006.
 #############################################################################
 
 package Graph::Easy::Layout::Grid;
 
 use vars qw/$VERSION/;
 
-$VERSION = '0.03';
+$VERSION = '0.04';
 
 #############################################################################
 #############################################################################
 
 package Graph::Easy;
 
+sub _balance_sizes
+  {
+  # Given a list of column/row sizes and a minimum size that their sum must
+  # be, will grow individual sizes until the constraint (sum) is met.
+  my ($self, $sizes, $need) = @_;
+
+  # XXX TODO: we can abort the loop and distribute the remaining nec. size
+  # once all elements in $sizes are equal.
+
+#  use Data::Dumper; print STDERR "# " . Dumper($sizes)," $need \n";
+
+  return if $need < 1;
+
+  # if there is only one element, return it immidiately
+  if (@$sizes == 1)
+    {
+    $sizes->[0] = $need if $sizes->[0] < $need;
+    return;
+    }
+
+  # endless loop until constraint is met
+  while (1)
+    {
+  
+    # find the smallest size, and also compute their sum
+    my $sum = 0; my $i = 0;
+    my $sm = $need + 1;		# start with an arbitrary size
+    my $sm_i = 0;		# if none is != 0, then use the first
+    for my $s (@$sizes)
+      {
+      $sum += $s;
+      next if $s == 0;
+      if ($s < $sm)
+	{
+        $sm = $s; $sm_i = $i; 
+	}
+      $i++;
+      }
+
+    # their sum is already equal or bigger than what we need?
+    last if $sum >= $need;
+
+    # increase the smallest size by one, then try again
+    $sizes->[$sm_i]++;
+    }
+ 
+#  use Data::Dumper; print STDERR "# " . Dumper($sizes),"\n";
+
+  undef;
+  }
+
 sub _prepare_layout
   {
   # this method is used by as_ascii() and as_svg() to find out the
@@ -39,15 +90,18 @@ sub _prepare_layout
   my $rows = {};
   my $cols = {};
 
-  # the last column/row
-  my $mx = -100000; my $my = -100000;
+  # the last column/row (highest X,Y pair)
+  my $mx = -1000000; my $my = -1000000;
+
+  # We need to do this twice, once for single-cell objects, and again for
+  # objects covering multiple cells. The single-cell objects can be solved
+  # first:
 
   # find all x and y occurances to sort them by row/columns
   for my $cell (values %$cells)
     {
     my ($x,$y) = ($cell->{x}, $cell->{y});
 
-    # Set the minimum cell size:
     {
       no strict 'refs';
 
@@ -56,13 +110,20 @@ sub _prepare_layout
       $cell->$method();
     }
 
-    my $w = $cell->{w};
-    my $h = $cell->{h};
+    my $w = $cell->{w} || 0;
+    my $h = $cell->{h} || 0;
 
-    # record maximum size for that col/row
-    $rows->{$y} = $h if $h >= ($rows->{$y} || 0);
-    $cols->{$x} = $w if $w >= ($cols->{$x} || 0);
+    # Set the minimum cell size only for single-celled objects:
+    if ( (($cell->{cx}||1) + ($cell->{cy}||1)) == 2)
+      { 
+      # record maximum size for that col/row
+      $rows->{$y} = $h if $h >= ($rows->{$y} || 0);
+      $cols->{$x} = $w if $w >= ($cols->{$x} || 0);
+      }
 
+    # Find highest X,Y pair. Always use x,y, and not x+cx,y+cy, because
+    # a multi-celled object "sticking" out will not count unless there
+    # is another object in the same row/column.
     $mx = $x if $x > $mx;
     $my = $y if $y > $my;
     } 
@@ -71,6 +132,63 @@ sub _prepare_layout
   $rows->{$my+1} = 0;
   $cols->{$mx+1} = 0;
 
+  # do the last step again, but for multi-celled objects
+  for my $cell (values %$cells)
+    {
+    my ($x,$y) = ($cell->{x}, $cell->{y});
+
+    my $w = $cell->{w} || 0;
+    my $h = $cell->{h} || 0;
+
+    # Set the minimum cell size only for multi-celled objects:
+    if ( (($cell->{cx} || 1) + ($cell->{cy}||1)) > 2)
+      {
+      $cell->{cx} ||= 1;
+      $cell->{cy} ||= 1;
+
+      # do this twice, for X and Y:
+
+#      print STDERR "\n# ", $cell->{name} || $cell->{id}, " cx=$cell->{cx} cy=$cell->{cy} $cell->{w},$cell->{h}:\n";
+
+      # create an array with the current sizes for the affacted rows/columns
+      my @sizes;
+
+#      print STDERR "# $cell->{cx} $cell->{cy} at cx:\n";
+
+      # XXX TODO: no need to do this for empty/zero cols
+      for (my $i = 0; $i < $cell->{cx}; $i++)
+        {
+        push @sizes, $cols->{$i+$x} || 0;
+	}
+      $self->_balance_sizes(\@sizes, $cell->{w});
+      # store the result back
+      for (my $i = 0; $i < $cell->{cx}; $i++)
+        {
+#        print STDERR "# store back $sizes[$i] to col ", $i+$x,"\n";
+        $cols->{$i+$x} = $sizes[$i];
+	}
+
+      @sizes = ();
+
+#      print STDERR "# $cell->{cx} $cell->{cy} at cy:\n";
+
+      # XXX TODO: no need to do this for empty/zero cols
+      for (my $i = 0; $i < $cell->{cy}; $i++)
+        {
+        push @sizes, $rows->{$i+$y} || 0;
+	}
+      $self->_balance_sizes(\@sizes, $cell->{h});
+      # store the result back
+      for (my $i = 0; $i < $cell->{cy}; $i++)
+        {
+#        print STDERR "# store back $sizes[$i] to row ", $i+$y,"\n";
+        $rows->{$i+$y} = $sizes[$i];
+	}
+	 
+      }
+
+    } 
+
   # Now run through all rows/columns and get their absolute pos by taking all
   # previous ones into account.
   my $pos = 0;
@@ -93,6 +211,7 @@ sub _prepare_layout
 
   for my $v (values %$cells)
     {
+    # Skip multi-celled nodes for later. 
     next if ($v->{cx}||1) + ($v->{cy}||1) != 2;
 
     # X and Y are col/row, so translate them to real pos
@@ -100,8 +219,8 @@ sub _prepare_layout
     my $y = $rows->{ $v->{y} };
 
     # Also set correct the width/height of each cell to be the maximum
-    # width/height of that colum and store the previous size in 'minw'
-    # and 'minh', respectively. Skip multi-celled nodes for later. 
+    # width/height of that row/column and store the previous size in 'minw'
+    # and 'minh', respectively.
 
     $v->{minw} = $v->{w};
     $v->{minh} = $v->{h};
@@ -124,9 +243,8 @@ sub _prepare_layout
     $max_x = $m if $m > $max_x;
     }
 
-  # repeat the previous step, now for multi-celled nodes
-  # Looping over nodes should be faster
-  foreach my $v (values %{$self->{nodes}})
+  # repeat the previous step, now for multi-celled objects
+  foreach my $v (values %{$self->{cells}})
     {
     next unless defined $v->{x} && (($v->{cx}||1) + ($v->{cy}||1) > 2);
 
@@ -138,9 +256,9 @@ sub _prepare_layout
     $v->{minh} = $v->{h};
 
     # find next col/row
-    my $nx = $v->{x} + $v->{cx};
+    my $nx = $v->{x} + ($v->{cx} || 1);
     my $next_col = $cols->{ $nx };
-    my $ny = $v->{y} + $v->{cy};
+    my $ny = $v->{y} + ($v->{cy} || 1);
     my $next_row = $rows->{ $ny };
 
     $next_col = $cols->{ ++$nx } while (!defined $next_col);
@@ -221,7 +339,7 @@ a list of all cells to be rendered.
 
 =head1 AUTHOR
 
-Copyright (C) 2004 - 2005 by Tels L<http://bloodgate.com>.
+Copyright (C) 2004 - 2006 by Tels L<http://bloodgate.com>.
 
 See the LICENSE file for information.
 
